@@ -1,11 +1,11 @@
 ;; -*- no-byte-compile: t; lexical-binding: nil -*-
-(define-package "tempel" "1.11"
+(define-package "tempel" "20251115.1817"
   "Tempo templates/snippets with in-buffer field editing."
   '((emacs  "29.1")
     (compat "30"))
   :url "https://github.com/minad/tempel"
-  :commit "8d9a1646fb4014a5d34f3f19b7c894310f06a1e8"
-  :revdesc "1.11-0-g8d9a1646fb40"
+  :commit "56b6e6e3e95054876cfb96b1f3a460824baec8c6"
+  :revdesc "56b6e6e3e950"
   :keywords '("abbrev" "languages" "tools" "text")
   :authors '(("Daniel Mendler" . "mail@daniel-mendler.de"))
   :maintainers '(("Daniel Mendler" . "mail@daniel-mendler.de")))
